@@ -1,7 +1,6 @@
 from hgraph._types._scalar_value import ScalarValue
 
-
-__all__ = ('PythonScalarValue', 'PythonKeyableScalarValue')
+__all__ = ("PythonScalarValue", "PythonKeyableScalarValue")
 
 
 class PythonScalarValue(ScalarValue):
@@ -46,4 +45,4 @@ class PythonKeyableScalarValue(PythonScalarValue):
     """
 
     def __hash__(self) -> int:
-        return hash(self._value)
\ No newline at end of file
+        return hash(self._value)
